----------------
-- File globals
----------------
local shortName = "SomeCooldowns"
local D = LibStub:GetLibrary("DoomCore-1.3")
local Addon = D.Addon(shortName, "Some Cooldowns")
Addon.version = 0.2

local K, nilSort, getItemID = D.K, D.nilSort, D.getItemID
local D_updateFrame, D_tooltip = D.updateFrame, D.tooltip

local inWorld
local gcd = 1.5
local frame
Addon.cooldowns = { spell = {}, item = {} }

local count = 0
--local showing = 0
local oldNumShown = {}
local loadDelay = 10
local sortDelay = 1
local abrupt = 3

--------
-- Core
--------
function Addon:OnInitialize()
  frame = CreateFrame("Frame", shortName, UIParent)
  frame:SetFrameStrata("LOW")
  frame.conf = self.core
  frame.tex = frame:CreateTexture()
  frame.tex:SetAllPoints(frame)
  frame.tex:SetColorTexture(0, 1, 0, 0.5)
  frame.tex:Hide()

  self:Register({
    media = "LibSharedMedia-3.0",
    masque = "Masque",
    somebars = "SomeBars"
  }, "Default")

  self:TrackEvent("PLAYER_ENTERING_WORLD")
  self:TrackEvent("GET_ITEM_INFO_RECEIVED")
  self:TrackMultiEventQueue("Update",
    "SPELL_UPDATE_COOLDOWN", "SPELL_UPDATE_USABLE", "SPELL_UPDATE_CHARGES",
    "BAG_UPDATE", "BAG_UPDATE_USABLE"
  )
end

function Addon:OnLoad()
  if self.lib.masque and not frame.msq then
    frame.msq = self.lib.masque:Group(self.name)
  end
  self:Draggable(frame, self.core)

  self.core.group = self.core.group or {}
  local settings = self.settings
  for val in self:GetAll(self.core.group, "blacklist") do
    settings:Add("blacklist", val)
  end
  for val in self:GetAll(self.core.group, "whitelist") do
    settings:Add("whitelist", val)
  end
  self:SetFilter()
end

function Addon:PLAYER_ENTERING_WORLD()
  inWorld = GetTime()
  self:OnLoad()
end

function Addon:GET_ITEM_INFO_RECEIVED()
  local crawler = self.settings.crawler
  for valName, val in pairs(self.core.group or {}) do
    if type(valName) == "number" then
      local entry = crawler:Get{ val.type, "args", tostring(valName) }
      if not entry and GetItemInfo(valName) then
        self:Add({ val.type }, valName)
      end
    end
  end
end

function Addon:Migrate()
  local oldVersion = self:GetVersion() or 0
  self:SetVersion(Addon.version)
  if oldVersion >= 0.2 then return end
  if self.core.rowGrowth == "up" then self.core.rowGrowth = "top" end
  if self.core.rowGrowth == "down" then self.core.rowGrowth = "bottom" end
end


function Addon:OnReset(registered)
  if registered == false and self:Migrate() then return end
end

function Addon:SetFilter(info, val)
  if not frame then return end
  local update = false
  if info then
    if self:ConfGet(info) ~= val then
      update = true
      self.cooldowns = { item = {}, spell = {} }
    end
    self:ConfSet(info, val)
  end
  for val in self:GetAll(self.core.group, "blacklist") do
    local cooldowns
    if type(val) == "number" then
      cooldowns = self.cooldowns.item
    else
      cooldowns = self.cooldowns.spell
    end
    cooldowns[val] = true
    update = true
  end

  self:Update(update)
end

function Addon:SetText(info, val)
  self:ConfSet(info, val)
  for _,v in pairs{ frame:GetChildren() } do
    v.cooldown:SetHideCountdownNumbers(not val)
  end
end

function Addon:SetReverse(info, val)
  self:ConfSet(info, val)
  for _,v in pairs{ frame:GetChildren() } do
    v.cooldown:SetReverse(val)
  end
end

function Addon:SetSwipeColor(info, r, g, b, a)
  self:ConfSet(info, r, g, b, a)
  for _,button in pairs{ frame:GetChildren() } do
    button.cooldown:SetSwipeColor(r, g, b, a)
  end
end


----------
-- Events
----------

function Addon:Sort()
  Addon:Output("Sorting")
  local children = { frame:GetChildren() }
  if #children == 0 then return end
  local remove = {}
  local add = {}
  local time = GetTime()
  --[[
  local suppressed,oldCount,oldTime = unpack(oldNumShown)
  oldCount = oldCount or 0
  self:Output(showing,time,"<-",suppressed,oldCount,math.floor(time - inWorld - loadDelay), oldTime and math.floor(time - oldTime))
  if suppressed then
    if showing < oldCount + abrupt then
      oldNumShown = { false, showing, time }
      frame:Show()
    end
  elseif showing > oldCount + abrupt and inWorld and time > inWorld + loadDelay then
    oldNumShown[1] = true
    frame:Hide()
  elseif not oldTime or time > oldTime + sortDelay then
    oldNumShown = { false, showing, time }
  end
  if showing == 0 then return end
  ]]--

  for _,button in pairs(children) do
    if button.active then
      local start,duration
      local type = button:GetAttribute("type")
      if type then
        local subject = button:GetAttribute(type)
        if subject then
          if type == "spell" then
            local charges, maxCharges, chargeStart, chargeDuration = GetSpellCharges(subject)
            if charges and charges < maxCharges and chargeStart and chargeStart > 0 and chargeDuration > gcd then
              start = chargeStart
              duration = chargeDuration * (maxCharges - charges)
            else
              start,duration = GetSpellCooldown(subject)
            end
          elseif type == "item" then
            start,duration = GetItemCooldown(subject)
          end
          if not start or not duration or duration <= gcd or start + duration + 1 <= time
              or not button.cooldown:IsShown()
              or button.cooldown:GetCooldownDuration() == 0 then
            tinsert(remove, button)
          else
            tinsert(add, button)
          end
        end
      end
    end
  end

  for _,button in pairs(remove) do
    button.active = false
    button:Hide()
  end

  if #add == 0 then return end

  nilSort(add, "-expires")

  local core = Addon.core
  D.makeGrid{
    frame        = frame,
    els          = add,
    parent       = UIParent,
    anchor       = core.anchor,
    x            = core.offsetX,
    y            = core.offsetY,
    size         = core.iconSize,
    padding      = core.padding,
    spacing      = core.spacing,
    grow         = core.grow,
    rowGrowth    = core.rowGrowth,
    columnGrowth = core.columnGrowth,
    max          = core.max,
    limit        = core.limit,
    background   = core.background,
    border       = core.border,
    edge         = core.edge
  }

  if frame.msq then frame.msq:ReSkin() end
end

function Addon:OnSpellCooldown(name, icon)
  if not name then return end
  local spellCooldowns = self.cooldowns.spell
  local record = 0
  local oldRecord, button
  local info = spellCooldowns[name]
  if info then
    if info == true then return end
    oldRecord, button = unpack(info)
  else
    info = {}
    spellCooldowns[name] = info
  end
  if oldRecord == true then return end
  local start, duration, cooldown
  local charges, maxCharges, chargeStart, chargeDuration = GetSpellCharges(name)
  if charges then
    record = charges
    if oldRecord == record then return end
    if not self.core.recharge and charges > 0 then return end
    if record == maxCharges then return end
    if oldRecord and button and record > oldRecord then
      ActionButton_ShowOverlayGlow(button)
      ActionButton_HideOverlayGlow(button)
    end
    if charges < maxCharges and chargeStart and chargeStart > 0 and chargeDuration > gcd then
      start = chargeStart
      duration = chargeDuration * (maxCharges - charges)
      cooldown = true
    end
  else
    if oldRecord == record then return end
    start, duration, cooldown = GetSpellCooldown(name)
  end
  if name and start and start > 0 and duration > gcd and cooldown then
    info[1] = record
    self:AddButton("spell", name, start, duration, icon)
    self:Output("Sorting with spell ",name)
    return true
  end
end
function Addon:OnItemCooldown(name)
  if not name then return end
  local itemCooldowns = self.cooldowns.item
  local oldRecord, button
  local info = itemCooldowns[name]
  if info then
    if info == true then return end
    oldRecord, button = unpack(info)
    if oldRecord then return end
  else
    info = {}
    itemCooldowns[name] = info
  end
  if oldRecord then return end
  local start, duration, cooldown = GetItemCooldown(name)
  if not start then return end

  if start > 0 and duration > gcd and cooldown then
    info[1] = 0
    self:AddButton("item", name, start, duration, GetItemIcon(name))
    self:Output("Sorting with item ",name)
    return true
  end
end

function Addon:Update(hard)
  gcd = select(2, GetSpellCooldown(61304)) or gcd
  local resort = false
  if C_LossOfControl.GetEventInfo(1) then return end
  if hard == true then
    self:Output("Hard reset")
    for _,button in pairs{ frame:GetChildren() } do
      button.active = false
      button:Hide()
    end
    self.cooldowns = { spell = {}, item = {}}
  end
  for item in self:GetAll(self.core.group, "whitelist") do
    if type(item) == "number" then
      local fromCooldown = self:OnItemCooldown(item)
      if fromCooldown then
        self:Output("Sorted from whitelist item ",item)
      end
      resort = fromCooldown or resort
    else
      local icon = select(3, GetSpellInfo(item))
      local fromCooldown = self:OnSpellCooldown(item, icon)
      if fromCooldown then
        self:Output("Sorted from whitelist spell ",item)
      end
      resort = fromCooldown or resort
    end
  end

  if self.core.displaySpells then
    local index = 1
    while true do
      local _,id = GetSpellBookItemInfo(index, "spell")
      if not id then break end
      local spell,_,icon = GetSpellInfo(id)
      spell = GetSpellInfo(spell) or spell
      local fromCooldown = self:OnSpellCooldown(spell, icon)
      if fromCooldown then
        self:Output("Sorted from spell ",spell)
      end
      resort = fromCooldown or resort
      index = index + 1
    end
  end

  if self.core.displayItems then
    for slot=1,19 do
      local item = GetInventoryItemID("player", slot)
      local fromCooldown = self:OnItemCooldown(item)
      if fromCooldown then
        self:Output("Sorted from equipped item ",item)
      end
      resort = fromCooldown or resort
    end
    for bag = 0,NUM_BAG_SLOTS do
      for slot = 1,GetContainerNumSlots(bag) do
        local item = GetContainerItemID(bag, slot)
        local fromCooldown = self:OnItemCooldown(item)
        if fromCooldown then
          self:Output("Sorted from inventory item ",item)
        end
        resort = fromCooldown or resort
      end
    end
  end

  if self.core.displayToys then
    for i=1,C_ToyBox.GetNumToys() do
      local toy = C_ToyBox.GetToyFromIndex(i)
      if PlayerHasToy(toy) then
        local fromCooldown = self:OnItemCooldown(toy)
        if fromCooldown then
          self:Output("Sorted from toy ",toy)
        end
        resort = fromCooldown or resort
      end
    end
  end
  if resort then self:Queue("Sort") end
end

----------
-- Frames
----------
function Addon:Rebuild(info, r, g, b, a)
  if info then self:ConfSet(info, r, g, b, a) end
  local size = self.core.iconSize
  frame:SetSize(size + 4, size + 4)

  for _,button in pairs{ frame:GetChildren() } do
    button:SetSize(size, size)
  end

  self:Update()
  self:Sort()

  D_updateFrame(frame, self.core, false)
end

function Addon:Add(info, val)
  local group
  if info[1] == "addToWhitelist" then
    group = "whitelist"
  elseif info[1] == "addToBlacklist" then
    group = "blacklist"
  else return end

  if not GetSpellInfo(val) then
    val = getItemID(val)
  end
  if not val then return end

  if info[1] == "addToBlacklist" then
    for _,b in pairs{ frame:GetChildren() } do
      local contained = b:GetAttribute(b:GetAttribute("type"))
      if contained == val then
        b.active = false
        b:Hide()
        break
      end
    end
  end

  self:Make(self.core.group, group, val)

  self.settings:Add(group, val)

  self:Update()
end

function Addon:InitButton(button)
  button:Hide()
  ActionButton_ShowOverlayGlow(button)
  ActionButton_HideOverlayGlow(button)
  for _,child in pairs{ button:GetChildren() } do
    if child.outerGlowOver then
      child.outerGlow:Hide()
      child.outerGlowOver:Hide()
      child.ants:Hide()
    end
  end
  button:SetAttribute("unit","player")
  button.cooldownFrame = _G[button:GetName() .. "Cooldown"]
  button.cooldownFrame:SetEdgeTexture("Interface\\Cooldown\\edge-LoC")
  button.cooldownFrame:SetScript("OnHide", self.Sort)
  button.cooldown:SetSwipeColor(self:ConfGet{ "color" })
  button.cooldown:SetHideCountdownNumbers(not self.core.text)
  button:Disable()
  D_tooltip(button, self.core.tooltip, self.core, function(button)
    local type = button:GetAttribute("type")
    if not type then return end
    local subject = button:GetAttribute(type)
    Addon:Output("Canceling ",type," ",subject)
    button.active = false
    Addon.cooldowns[type][subject] = {}
  end)
end

function Addon:AddButton(buttonType, subject, time, duration, texture)
  local typeCooldowns = self.cooldowns[buttonType]
  for blacklisted in self:GetAll(self.core.group, "blacklist") do
    if subject == blacklisted
        or (buttonType == "spell" and GetSpellInfo(subject) == blacklisted)
        or (buttonType == "item" and GetItemInfo(subject) == GetItemInfo(blacklisted)) then
      return
    end
  end
  if self:ConfGetDebug{ "somebars" } then
    for frameName, alsoName, also in self:GetAll(self.lib.somebars, "also", 3) do
      if subject == alsoName
          or (buttonType == "spell" and GetSpellInfo(subject) == alsoName)
          or (buttonType == "item" and GetItemInfo(subject) == GetItemInfo(alsoName)) then
        return
      end
    end
  end
  local button
  local oldInfo = typeCooldowns[subject]
  if oldInfo then
    button = oldInfo[2]
  else
    oldInfo = {}
    typeCooldowns[subject] = oldInfo
  end
  if not button then
    for _,b in pairs{ frame:GetChildren() } do
      if not b.active then
        button = b
        break
      end
    end
  end
  if not button then
    count = count + 1
    button = CreateFrame("Button", frame:GetName() .. count, frame, "ActionButtonTemplate")
    self:InitButton(button)
    if frame.msq then frame.msq:AddButton(button) end
  end
  typeCooldowns[subject][2] = button
  button.icon:SetTexture(texture)
  local iconSize = self.core.iconSize
  button:SetSize(iconSize, iconSize)
  button:SetAttribute("type", buttonType)
  button:SetAttribute(buttonType, subject)
  button:GetNormalTexture():Hide()
  button.cooldownFrame:SetCooldown(time, duration)
  button.cooldownFrame:SetReverse(self.core.reverse)
  button.expires = time + duration
  button.active = true
  --button:Show()
end

function Addon:UpdateTooltips(info, val)
  self:ConfSet(info, val)
  local tooltip = self.core.tooltip
  local override = self.core.tooltipOverride
  local anchor = self.core.tooltipAnchor
  for _,button in pairs{ frame:GetChildren() } do
    if button:IsMouseOver() then
      if not tooltip then button:GetScript("OnLeave")(button)
      elseif override then
        GameTooltip:SetOwner(button,anchor)
      else
        GameTooltip:SetOwner(button)
      end
    end
    button:SetMotionScriptsWhileDisabled(tooltip)
  end
end